/*
  dsniff.c

  simple libnids-based sniffer, because DrHoney wanted one.

  minimally decodes FTP, Telnet, HTTP, POP, IMAP, SNMP, NFS auth info.
  this is for demonstration purposes and educational use only.
  
  Copyright (c) 1999 Dug Song <dugsong@monkey.org>
  All rights reserved, all wrongs reversed.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. The name of author may not be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  $Id: dsniff.c,v 1.11 2000/01/04 20:26:53 dugsong Exp $
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <arpa/telnet.h>
#include <rpc/rpc.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <nids.h>

#include "version.h"

#define MAX_LINES	6
#define MIN_SNAPLEN	512
#define PMAPSIZE	64
#define XIDMAPSIZE	64
#define PROG_PMAP	100000
#define PROG_MOUNT	100005

#define int_ntoa(x)	inet_ntoa(*((struct in_addr *)&(x)))

#define FRAGLEN(x)	(x & 0x7fffffff)
#define LASTFRAG(x)	(x & (1 << 31))

struct pmap_entry {
  u_long dst;
  u_short dport;
  int prog;
};

struct xid_map_entry {
  u_long xid;
  u_long src;
  u_long dst;
  int prog;
  char *data;
};

struct rpc_record {
  u_long rm;
  u_char *buf;
  int len;
};

struct rpc_call {
  int xid;
  int prog;
  int proc;
  u_char *cred;
  u_char *verf;
  u_char *args;
};

struct rpc_reply { /* accepted, successful only */
  int xid;
  u_char *verf;
  u_char *results;
};

extern int base64_decode(char *in, char *out, int len);

/* Globals. */
char	Buf[BUFSIZ];
int	Lines = MAX_LINES;
int	Snaplen = MIN_SNAPLEN;
int	pmap_next = 0;
int	pmap_hint = 0;
int	xid_map_next = 0;
int	xid_map_hint = 0;
struct pmap_entry pmap_map[PMAPSIZE];
struct xid_map_entry xid_map[XIDMAPSIZE];

void
usage(void)
{
  fprintf(stderr, "Usage: dsniff [-i interface]\n");
  exit(1);
}

#ifndef OpenBSD
size_t
strlcpy(char *dst, const char *src, size_t siz)
{
  char *d = dst;
  const char *s = src;
  size_t n = siz;
  
  if (n != 0 && --n != 0) {
    do {
      if ((*d++ = *s++) == 0) break;
    } while (--n != 0);
  }
  if (n == 0) {
    if (siz != 0) *d = '\0';
    while (*s++) ;
  }
  return (s - src - 1);
}

size_t
strlcat(char *dst, const char *src, size_t siz)
{
  char *d = dst;
  const char *s = src;
  size_t dlen, n = siz;

  while (*d != '\0' && n-- != 0) d++;
  dlen = d - dst;
  n = siz - dlen;
  
  if (n == 0) return (dlen + strlen(s));
  for ( ; *s != '\0'; s++) {
    if (n != 1) {
      *d++ = *s;
      n--;
    }
  }
  *d = '\0';
  
  return (dlen + (s - src));
}
#endif /* !OpenBSD */

#ifdef DEBUG
/* Print a buffer in hex. */
void
hex_print(u_char *buf, int len)
{
  int i;
  
  for (i = 0; i < len; i++) {
    if (!(i & 0xF)) printf("\n");
    else if (!(i % 2)) printf(" ");
    printf("%.2x", buf[i]);
  }
  printf("\n");
}
#endif /* DEBUG */

/* Locate substring in a binary string. */
u_char *
bufstr(u_char *buf1, int len1, u_char *buf2, int len2)
{
  u_char *p;

  for (p = buf1; p < buf1 + len1 - len2; p++) {
    if (memcmp(p, buf2, len2) == 0)
      return (p);
  }
  return (NULL);
}
  
/* Strip telnet options, as well as suboption data. */
int
strip_telopts(u_char *buf, int len)
{
  int i, j, subopt = 0;
  char *p, *q;
  
  for (i = j = 0; i < len; i++) {
    if (buf[i] == IAC) {
      if (++i >= len) break;
      else if (buf[i] > SB)
	i++;
      else if (buf[i] == SB) {
	/* XXX - save autologin username, if any. */
	p = buf + i + 1;
	if ((q = bufstr(p, len - i, "\xff", 1)) != NULL) {
	  if ((p = bufstr(p, q - p, "USER\x01", 5)) != NULL) {
	    p += 5;
	    buf[j++] = '[';
	    memcpy(buf + j, p, q - p); j += q - p;
	    buf[j++] = ']'; buf[j++] = '\n';
	  }
	}
	subopt = 1;
      }
      else if (buf[i] == SE) {
	if (!subopt) j = 0;
	subopt = 0;
      }
    }
    else if (!subopt) {
      /* XXX - convert isolated carriage returns to newlines. */
      if (buf[i] == '\r' && i + 1 < len && buf[i + 1] != '\n')
	buf[j++] = '\n';
      else
	buf[j++] = buf[i];
    }
  }
  buf[j] = '\0';
  
  return (j);
}

/* Strip a buffer down to a maximum number of lines. */
int
strip_lines(char *buf, int max_lines)
{
  int num = 0;

  if (!buf) return (0);

  while (num < max_lines && (buf = strchr(buf, '\n')) != NULL) {
    buf++; num++;
  }
  if (buf) *buf = '\0';

  return (num);
}

int
decode_ftp_auth(char *buf, int len)
{
  char *p;
  
  Buf[0] = '\0';

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if (strncasecmp(p, "USER ", 5) == 0 ||
	strncasecmp(p, "PASS ", 5) == 0 ||
	strncasecmp(p, "ACCT ", 5) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
  }
  return (strlen(Buf));
}

int
decode_http_auth(char *buf, int len)
{
  char *p, *q;
  int i;

  /* Chop header. */
  if ((p = strstr(buf, "\r\n\r\n")) == NULL)
    if ((p = strstr(buf, "\n\n")) == NULL)
      return (0);

  *p = Buf[0] = '\0';

  /* Check for auth info. */
  if ((p = strstr(buf, "uthorization: ")) == NULL)
    return (0);

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if (!strlen(p)) {
      continue;
    }
    else if (strncasecmp(p, "GET ", 4) == 0 ||
	strncasecmp(p, "POST ", 5) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
    else if (strncasecmp(p, "Host: ", 6) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
    else if (strncasecmp(p, "Authorization: Basic ", 21) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, " [", sizeof(Buf));
      p += 21;
      i = strlen(Buf);
      q = Buf + i;
      q += base64_decode(p, q, len - i - 2);
      q[0] = '\0';
      strlcat(Buf, "]\n", sizeof(Buf));
    }
  }
  return (strlen(Buf));
}

int
decode_pop_auth(char *buf, int len)
{
  char *p, *q;
  int i;
  
  Buf[0] = '\0';

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if (strncasecmp(p, "AUTH PLAIN", 10) == 0 ||
	strncasecmp(p, "AUTH LOGIN", 10) == 0) {
      char *user, *pass;
      
      /* Decode SASL auth. */
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));

      if ((user = strtok(NULL, "\r\n")) != NULL &&
	  (pass = strtok(NULL, "\r\n")) != NULL) {
	strlcat(Buf, user, sizeof(Buf));
	strlcat(Buf, "\n", sizeof(Buf));
	strlcat(Buf, pass, sizeof(Buf));
	strlcat(Buf, " [", sizeof(Buf));
	i = strlen(Buf);
	q = Buf + i;
	q += base64_decode(pass, q, len - i - 2);
	*q = '\0';
	strlcat(Buf, "]\n", sizeof(Buf));
      }
    }
    /* Save regular POP2, POP3 auth info. */
    else if (strncasecmp(p, "USER ", 5) == 0 ||
	     strncasecmp(p, "PASS ", 5) == 0 ||
	     strncasecmp(p, "HELO ", 5) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
  }
  return (strlen(Buf));
}

int
decode_imap_auth(char *buf, int len)
{
  char *p, *q;
  
  Buf[0] = '\0';

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if ((q = strchr(p, ' ')) != NULL) {
      if (strncasecmp(q++, "LOGIN ", 6) == 0) {
	strlcat(Buf, p, sizeof(Buf));
	strlcat(Buf, "\n", sizeof(Buf));
      }
    }
  }
  return (strlen(Buf));
}

int
decode_rlogin_auth(char *buf, int len)
{
  char *p, *q;

  p = buf + 1;				/* Skip first NULL */

  strlcpy(Buf, "[", sizeof(Buf));
  strlcat(Buf, p, sizeof(Buf));		/* Local username */
  strlcat(Buf, ":", sizeof(Buf));
  p += strlen(p) + 1;

  strlcat(Buf, p, sizeof(Buf));		/* Remote username */
  strlcat(Buf, "]\n", sizeof(Buf));
  p += strlen(p) + 1;

  p += strlen(p) + 1;			/* Skip term info */

  if ((q = strstr(p, "\xff\xffss")) != NULL)	/* Skip window size */
    p += 12;
  
  for (p = strtok(p, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    strlcat(Buf, p, sizeof(Buf));
    strlcat(Buf, "\n", sizeof(Buf));
  }
  return (strlen(Buf));
}

int
get_asn1_len(u_char **buf)
{
  u_char *p;
  int len, num = 0;
  
  p = *buf;
  
  if (*p >= 128) {	/* Long form */
    len = *p++ & ~128;
    
    if (len == 1) {
      num = *p++;
    }
    else if (len == 2) {
      GETSHORT(num, p);
    }
    else if (len == 3) {
      p--; GETLONG(num, p);
      num &= 0xFFF;
    }
    else if (len== 4)
      GETLONG(num, p);
  }
  else num = *p++;	/* Short form */

  *buf = p;
  
  return (num);
}

int
decode_snmp_auth(u_char *buf, int len)
{
  u_char *p = buf;
  int i, vers;
  
  if ((*p++ & 0x1F) != 16)	/* XXX - check for sequence */
    return (0);
  i = get_asn1_len(&p);	/* XXX - skip sequence length */
  
  if ((*p++ & 0x1F) != 2)	/* XXX - check for int */
    return (0);
  if (get_asn1_len(&p) != 1)	/* XXX - check version length (always 1?) */
    return (0);
  vers = *p++;

  p++;				/* XXX - skip class, tag. */
  i = get_asn1_len(&p);		/* community string length */
  p[i] = '\0';

  snprintf(Buf, sizeof(Buf), "[version %d]\n%s\n", vers + 1, p);
  
  return (strlen(Buf));
}

/* XXX - seriously lame portmap cache. ugh. */
void
pmap_enter(u_long dst, u_short dport)
{
  struct pmap_entry *pp;

  pp = &pmap_map[pmap_next];
  pp->dst = dst;
  pp->dport = dport;

  if (++pmap_next >= PMAPSIZE)
    pmap_next = 0;
}

int
pmap_find(u_long dst, u_short dport)
{
  struct pmap_entry *pp;
  int i;
  
  if (dport == 111)
    return (1);
  
  i = pmap_hint;
  do {
    pp = &pmap_map[i];
    if (pp->dst == dst && pp->dport == dport) {
      /* match */
      pmap_hint = i;
      return (1);
    }
    if (++i >= PMAPSIZE)
      i = 0;
  } while (i != pmap_hint);
  
  return (0);
}

/* xid_map borrowed from tcpdump's print-nfs.c */
void
xid_map_enter(u_long src, u_long dst, int xid, int prog, u_char *data)
{
  struct xid_map_entry *mp;
  
  mp = &xid_map[xid_map_next];
  
  if (++xid_map_next >= XIDMAPSIZE)
    xid_map_next = 0;
  
  mp->xid = xid;
  mp->src = src;
  mp->dst = dst;
  mp->prog = prog;
  mp->data = data;
}

struct xid_map_entry *
xid_map_find(u_long src, u_long dst, int xid)
{
  struct xid_map_entry *mp;
  int i;

  /* Start searching from where we last left off. */
  i = xid_map_hint;
  do {
    mp = &xid_map[i];
    if (mp->xid == xid && mp->src == src && mp->dst == dst) {
      /* match */
      xid_map_hint = i;
      return (mp);
    }
    if (++i >= XIDMAPSIZE)
      i = 0;
  } while (i != xid_map_hint);
  
  return (NULL);
}

/* XXX - Minimal RPC decoding. We don't even verify lengths, etc. */
int
decode_rpc(struct tuple4 *addr, u_char *buf, int len)
{
  struct rpc_msg *rm;
  struct rpc_call rc;
  struct rpc_reply rr;
  struct xid_map_entry *xp;
  char fh[96];
  u_char *p;
  int i;
  
  rm = (struct rpc_msg *)buf;

  if (pmap_find(addr->daddr, addr->dest) && ntohl(rm->rm_direction) == CALL) {
    rc.xid = ntohl(rm->rm_xid);
    rc.prog = ntohl(rm->rm_call.cb_prog);
    rc.proc = ntohl(rm->rm_call.cb_proc);
    rc.cred = (u_char *)&rm->rm_call.cb_cred;
    p = rc.cred + 4;				/* skip cred type */
    GETLONG(i, p); p += i;			/* skip cred data */
    rc.verf = p; p += 4;			/* skip verf type */
    GETLONG(i, p); p += i;			/* skip verf data */
    rc.args = p;

    switch (rc.prog) {
    case PROG_MOUNT:
      if (rc.proc != 1) return (0);
      i = ntohl(*((u_long *)rc.args));
      if ((p = malloc(i + 1)) == NULL)
	nids_params.no_mem("decode_rpc");
      strlcpy(p, rc.args + 4, i + 1);
      xid_map_enter(addr->saddr, addr->daddr, rc.xid, rc.prog, p);
      break;
    case PROG_PMAP:
      if (rc.proc != 3) return (0);
      if ((p = malloc(sizeof(u_long))) == NULL)
	nids_params.no_mem("decode_rpc");
      if (ntohl(*((u_long *)rc.args)) == PROG_MOUNT)
	xid_map_enter(addr->saddr, addr->daddr, rc.xid, rc.prog, NULL);
      break;
    }
  }
  else if (pmap_find(addr->saddr, addr->source) &&
	   ntohl(rm->rm_direction) == REPLY &&
	   ntohl(rm->rm_reply.rp_stat) == MSG_ACCEPTED) {
    rr.xid = ntohl(rm->rm_xid);
    rr.verf = (u_char *)&rm->rm_reply.rp_acpt.ar_verf;
    p = rr.verf + 4;				/* skip verf type */
    GETLONG(i, p); p += i;			/* skip verf data */
    GETLONG(i, p);				/* get accept_stat */
    if (i != SUCCESS) return (0);
    rr.results = p;
    
    if ((xp = xid_map_find(addr->daddr, addr->saddr, rr.xid)) == NULL)
      return (0);

    switch (xp->prog) {
    case PROG_MOUNT:
      for (i = 0; i < 32; i++)
	sprintf(fh + (i * 3), "%02x ", rr.results[i]);
      fh[95] = '\0';
      snprintf(Buf, sizeof(Buf), "%s [%s]\n", xp->data, fh);
      free(xp->data);
      memset(xp, 0, sizeof(*xp));
      return (1);
      break;
    case PROG_PMAP:
      pmap_enter(addr->saddr, ntohl(*((u_long *)rr.results)));
      break;
    }
  }
  return (0);
}

void
print_session(u_long src, u_short sport,
	      u_long dst, u_short dport, char *service)
{
  time_t t;
  struct tm *tm;
  char now[24];

  t = time(NULL);
  tm = localtime(&t);
  strftime(now, sizeof(now), "%x %X", tm);
  
  printf("---------------------------------\n");
  printf("%s %s:%d -> ", now, int_ntoa(src), sport);
  printf("%s:%d (%s)\n", int_ntoa(dst), dport, service);
  printf("%s\n", Buf);
  
  fflush(stdout);
}

void
print_udp_client(struct ip *ip, struct udphdr *udp)
{
  u_char *udp_b;
  int udp_bl;
  char *service;

  udp_b = (u_char *)(udp + 1);
  udp_bl = ntohs(udp->uh_ulen) - sizeof(*udp);
  
  switch (ntohs(udp->uh_dport)) {
  case 161:
    service = "snmp";
    if (!decode_snmp_auth(udp_b, udp_bl))
      return;
    break;
  default:
    return;
    break;
  }
  print_session(ip->ip_src.s_addr, ntohs(udp->uh_sport),
		ip->ip_dst.s_addr, ntohs(udp->uh_dport), service);
}
  
void
print_tcp_client(struct tcp_stream *ts, int len)
{
  char *service;
  
  if (len <= 0) return;

  /* Make subsequent string operations safe. */
  if (ts->server.bufsize > len)
    ts->server.data[len] = '\0';
  
  /* Pre-process output. */
  switch (ts->addr.dest) {
  case 21:
    service = "ftp";
    if (strip_telopts(ts->server.data, len) == 0)
      return;
    if (!decode_ftp_auth(ts->server.data, len))
      return;
    break;
  case 23:
    service = "telnet";
    if (strip_telopts(ts->server.data, len) == 0)
      return;
    if (strip_lines(ts->server.data, Lines) == 0)
      return;
    break;
  case 80:
    service = "http";
    if (!decode_http_auth(ts->server.data, len))
      return;
    break;
  case 109:
  case 110:
    service = "pop";
    if (!decode_pop_auth(ts->server.data, len))
      return;
    break;
  case 143:
    service = "imap";
    if (!decode_imap_auth(ts->server.data, len))
      return;
    break;
  case 513:
    service = "rlogin";
    if (!decode_rlogin_auth(ts->server.data, len))
      return;
    if (strip_lines(ts->server.data, Lines) == 0)
      return;
    break;
  default:
    service = "unknown";
    if (strip_lines(ts->server.data, Lines) == 0)
      return;
    break;
  }
  print_session(ts->addr.saddr, ts->addr.source,
		ts->addr.daddr, ts->addr.dest, service);
}

void
sniff_udp_client(struct ip *ip)
{
  struct udphdr *udp;
  int ip_hl = ip->ip_hl * 4;
  int ip_blen;
  
  if (ip->ip_p != IPPROTO_UDP || ntohs(ip->ip_len) - ip_hl < sizeof(*udp))
    return;
  
  ip_blen = ntohs(ip->ip_len) - ip_hl;
  udp = (struct udphdr *)((u_char *)ip + ip_hl);
  
  if (ip_blen != ntohs(udp->uh_ulen))
    return;
  
  switch (ntohs(udp->uh_dport)) {
  case 161:	/* SNMP */
    break;
  default:
    return;
    break;
  }
  print_udp_client(ip, udp);
}

void
sniff_tcp_client(struct tcp_stream *ts, void **whatever)
{
  int len;

  switch (ts->addr.dest) {
  case 21:	/* FTP */
  case 23:	/* Telnet */
  case 80:	/* HTTP */
  case 109:	/* POP */
  case 110:
  case 143:	/* IMAP */
  case 220:
  case 513:	/* Rlogin */
    break;
  default:
    return;
    break;
  }  
  switch (ts->nids_state) {
    
  case NIDS_JUST_EST:
    ts->server.collect = 1;
    break;
    
  case NIDS_DATA:
    len = ts->server.count - ts->server.offset;

    if (len < MIN_SNAPLEN) {
      /* Not enough data yet, save buffer. */
      nids_discard (ts, 0);
    }
    else {
      /* Log and discard this stream. */
      print_tcp_client(ts, len);
      ts->server.collect = 0;
    }
    break;
    
  default:
    /* Connection closing, log what we can. */
    print_tcp_client(ts, ts->server.count - ts->server.offset);
    ts->server.collect = 0;
    break;
  }
}

void
sniff_udp_rpc(struct ip *ip)
{
  struct udphdr *udp;
  int ip_hl = ip->ip_hl * 4;
  int ip_blen, udp_blen;
  struct tuple4 addr;

  /* XXX - basic UDP validation. need to implement nids_register_udp(). */
  if (ip->ip_p != IPPROTO_UDP || ntohs(ip->ip_len) - ip_hl < sizeof(*udp))
    return;
  
  ip_blen = ntohs(ip->ip_len) - ip_hl;
  udp = (struct udphdr *)((u_char *)ip + ip_hl);
  
  if (ip_blen != ntohs(udp->uh_ulen)) /* XXX - truncated? */
    return;
  
  udp_blen = ip_blen - sizeof(*udp);
  
  addr.saddr = ip->ip_src.s_addr;
  addr.daddr = ip->ip_dst.s_addr;
  addr.source = ntohs(udp->uh_sport);
  addr.dest = ntohs(udp->uh_dport);

  if (!pmap_find(addr.daddr, addr.dest) &&
      !pmap_find(addr.saddr, addr.source))
    return;
  
  if (!decode_rpc(&addr, (u_char *)(udp + 1), ip_blen - sizeof(*udp)))
    return;
  
  /* XXX - only mountd for now. */
  print_session(ip->ip_dst.s_addr, ntohs(udp->uh_dport),
		ip->ip_src.s_addr, ntohs(udp->uh_sport), "mountd");
}

void
sniff_tcp_rpc(struct tcp_stream *ts, void **conn_save)
{
  struct rpc_record *rr;
  struct tuple4 msgaddr;
  u_char *buf;
  int i;

  if (!pmap_find(ts->addr.daddr, ts->addr.dest))
    return;

  switch (ts->nids_state) {
    
  case NIDS_JUST_EST:
    ts->client.collect = ts->server.collect = 1;
    if ((rr = calloc(sizeof(*rr) * 2, 1)) == NULL)
      nids_params.no_mem("sniff_tcp_rpc");
    *conn_save = (void *)rr;
    break;
    
  case NIDS_DATA:
    if (ts->server.count_new) {
      rr = (struct rpc_record *)*conn_save;
      buf = ts->server.data;
      i = ts->server.count;
      msgaddr = ts->addr;
    }
    else {
      rr = (struct rpc_record *)*conn_save + 1;
      buf = ts->client.data;
      i = ts->client.count;
      msgaddr.saddr = ts->addr.daddr;
      msgaddr.daddr = ts->addr.saddr;
      msgaddr.source = ts->addr.dest;
      msgaddr.dest = ts->addr.source;
    }
    
    if (rr->rm == 0) {
      rr->rm = ntohl(*((u_long *)buf));
    }
    /* Check for complete fragment. */
    if (i < 4 + FRAGLEN(rr->rm)) {
      nids_discard(ts, 0);
      return;
    }
    /* We have a complete fragment. Build up message. */
    if (rr->buf == NULL && (rr->buf = malloc(FRAGLEN(rr->rm))) == NULL)
      nids_params.no_mem("sniff_tcp_rpc");
    else if ((rr->buf = realloc(rr->buf, rr->len + FRAGLEN(rr->rm))) == NULL)
      nids_params.no_mem("sniff_tcp_rpc");
    
    memcpy(rr->buf + rr->len, buf + 4, FRAGLEN(rr->rm));
    rr->len += FRAGLEN(rr->rm);

    /* Check for complete message. */
    if (!LASTFRAG(rr->rm)) {
      rr->rm = 0;
      nids_discard(ts, 4 + FRAGLEN(rr->rm));
      return;
    }
    rr->rm = 0;
    
    /* XXX - only mountd for now. */
    if (decode_rpc(&msgaddr, rr->buf, rr->len)) {
      print_session(ts->addr.saddr, ts->addr.source,
		    ts->addr.daddr, ts->addr.dest, "mountd");
    }
    break;
    
  default:
    rr = (struct rpc_record *)*conn_save;
    for (i = 0; i < 2; i++) {
      if (rr[i].buf != NULL)
	free(rr[i].buf);
    }
    free(rr);
    ts->client.collect = ts->server.collect = 0;
    break;
  }
}

int
main(int argc, char *argv[])
{
  char c;

  while ((c = getopt(argc, argv, "i:h?V")) != EOF) {
    switch (c) {
    case 'i':
      nids_params.device = optarg;
      break;
    case 'V':
      fprintf(stderr, "Version: %s\n", VERSION);
      usage();
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0)
    usage();
  
  memset(&pmap_map, 0, sizeof(pmap_map));
  memset(&xid_map, 0, sizeof(xid_map));
  
  /*
    this will have to wait until the new libnids version is out.
    nids_params.scan_num_hosts = 0;
  */

  if (!nids_init()) {
    fprintf (stderr, "%s\n", nids_errbuf);
    exit(1);
  }
  nids_register_ip(sniff_udp_client);
  nids_register_ip(sniff_udp_rpc);
  nids_register_tcp(sniff_tcp_client);
  nids_register_tcp(sniff_tcp_rpc);
  
  nids_run();

  /* NOTREACHED */
  exit(0);
}

/* 5000. */
